from typing import Union, Tuple, Optional, Sequence
import numpy as np
from ase import Atoms
from ase.build import surface, make_supercell
from ase.geometry import get_layers

from .concentration import Concentration
from .settings import ClusterExpansionSettings

__all__ = ('CESlab',)


def CESlab(conventional_cell: Union[Atoms, str],
           miller: Tuple[int],
           concentration: Concentration,
           size: Optional[Sequence[int]] = (1, 1, 1),
           **kwargs) -> ClusterExpansionSettings:
    """

    :param conventional_cell:
        Bulk lattice structure. Note that the unit-cell
        must be the conventional cell - not the primitive cell. One can also
        give the chemical symbol as a string, in which case the correct bulk
        lattice will be generated automatically.

    :param miller:
        Surface normal in Miller indices (h,k,l).

    :param concentration:
        Class for restricting the concentrations

    :param size:
        Size of the simulations cell. The third number represents the number of
        layers. The two first are repetitions of the in-plane unit vectors

    For more kwargs, see docstring of :class:`clease.settings.ClusterExpansionSettings`.
    """

    for b in concentration.basis_elements:
        if 'X' not in b:
            raise ValueError("Slab calculation requires that X is present in all basis")

    prim = get_prim_slab_cell(conventional_cell, miller)

    # Slab should always have one cell vector along the z-axis
    settings = ClusterExpansionSettings(prim, concentration, size=size, **kwargs)

    dict_rep = conventional_cell.todict()
    for k, v in dict_rep.items():
        if isinstance(v, np.ndarray):
            dict_rep[k] = v.tolist()

    settings.kwargs.update({
        'factory': 'CESlab',
        'miller': miller,
        'conventional_cell': dict_rep,
        'size': size
    })
    return settings


def get_prim_slab_cell(conventional_cell: Union[Atoms, str], miller: Tuple[int]) -> Atoms:
    """
    Returns the primitive cell used for slab CE

    :param conventional_cell:
        Bulk lattice structure. Note that the unit cell must be a conventional
        cell - not the primitive cell. One can also give the chemical symbol
        as a string, in which case the correct bulk lattice will be generated
        automatically.

    :param miller:
        Surface normal in Miller indices (h,k,l).
    """
    prim = surface(conventional_cell, miller, 1, periodic=True)
    return prim


def add_vacuum_layers(atoms: Atoms, prim: Atoms, thickness: float) -> Atoms:
    """
    Add vacuum layers to the slab

    :param atoms:
        ASE Atoms object representing the slab

    :param thickness:
        Thickness of the vacuum layer
    """
    # construct one layer of vacuum filled with vacancies
    prim_xy = prim.cell[:3]
    atoms_xy = atoms.cell[:3]
    P = atoms_xy.dot(np.linalg.inv(prim_xy))[:2].astype(int)
    P = np.vstack((P, [0, 0, 1]))
    vacuum = make_supercell(prim, P)
    for atom in vacuum:
        atom.symbol = 'X'

    dz = atoms.get_cell()[2, 2]
    vacuum.translate([0, 0, dz])

    one_layer_thickness = vacuum.get_cell()[2, 2]
    # upside-down division to get ceiling division
    num_vac_layers = int(-(-thickness // one_layer_thickness))
    vacuum *= (1, 1, num_vac_layers)

    atoms += vacuum
    atoms.cell[2, 2] += vacuum.cell[2, 2]
    return atoms


def remove_vacuum_layers(atoms: Atoms) -> Atoms:
    """
    Remove vacuum layers from the slab.

    :param atoms:
        ASE Atoms object representing the slab with vacuum
    """
    atoms = atoms.copy()
    tags, _ = get_layers(atoms, miller=(0, 0, 1))
    num_total_layers = np.amax(tags, axis=None) + 1
    vac_layers = []

    for layer in range(num_total_layers):
        symbols = [atoms[i].symbol for i, tag in enumerate(tags) if tag == layer]
        if set(symbols) == {'X'}:
            vac_layers.append(layer)

    del atoms[[atoms[i].index for i, tag in enumerate(tags) if tag in vac_layers]]

    num_vac_layers = len(vac_layers)
    atoms.cell[2, 2] *= (num_total_layers - num_vac_layers) / num_total_layers

    return atoms



from typing import Sequence, Dict, Any, List
from functools import total_ordering
import numpy as np
from ase import Atoms
import attr
from clease.datastructures import Figure
from clease.tools import equivalent_deco, list2str
from .cluster_fingerprint import ClusterFingerprint

__all__ = ('Cluster',)


@total_ordering
@attr.s(order=False, eq=False)
class Cluster:
    """A Cluster class, which collects multiple symmetry equivalent Figure objects,
    and related properties."""

    name: str = attr.ib()
    size: int = attr.ib()
    diameter: float = attr.ib()
    fingerprint: ClusterFingerprint = attr.ib(
        validator=attr.validators.instance_of(ClusterFingerprint))
    figures: Sequence[Figure] = attr.ib()
    equiv_sites: Sequence[Sequence[int]] = attr.ib()
    group: int = attr.ib()

    info: Dict[str, Any] = attr.ib(default=attr.Factory(dict))
    # "indices" are the integer index representation of the Figures.
    # therefore, "indices" and "ref_indx" depend on the currently active template,
    # and are subject to mutation.
    indices: Sequence[Sequence[int]] = attr.ib(default=None)
    ref_indx: int = attr.ib(default=None)

    @figures.validator
    def _validate_figures(self, attribute, value):
        """Verify that we have the correct type in the "figures" field."""
        # pylint: disable=unused-argument, no-self-use
        for ii, v in enumerate(value):
            if not isinstance(v, Figure):
                raise TypeError(f'All values must Figure type, got {value} '
                                f'of type {type(v)} in index {ii}.')

    @property
    def fp(self) -> ClusterFingerprint:
        """Alias for fingerprint, for compatibility."""
        return self.fingerprint

    def __lt__(self, other: 'Cluster') -> bool:
        """Less-than comparison operator."""
        if not isinstance(other, Cluster):
            return NotImplemented
        return self.fingerprint < other.fingerprint

    def __eq__(self, other: 'Cluster') -> bool:
        """Equals comparison operator."""
        if not isinstance(other, Cluster):
            return NotImplemented
        return self.fingerprint == other.fingerprint

    def equiv_deco(self, deco):
        return equivalent_deco(deco, self.equiv_sites)

    def todict(self) -> Dict[str, Any]:
        return attr.asdict(self)

    def is_subcluster(self, other) -> bool:
        """Check if the passed cluster is a subcluster of the current."""
        if len(self.indices) == 0:
            return True

        if len(self.indices[0]) >= len(other.indices[0]):
            return False

        return any(set(s1).issubset(s2) for s1 in self.indices for s2 in other.indices)

    def get_figure(self, primitive: Atoms, index: int = 0) -> Atoms:
        """Get figure from a ClusterGenerator object"""
        figure_four_vec = self.figures[index]
        positions = np.array([fv.to_cartesian(primitive) for fv in figure_four_vec])
        positions -= np.mean(positions, axis=0)
        symbols = [primitive[fv.sublattice] for fv in figure_four_vec]
        return Atoms(symbols, positions=positions)

    def get_figure_key(self, figure: Sequence[int]) -> str:
        """Return a key representation of the figure (in index representation)."""
        return list2str(self._order_equiv_sites(figure))

    def _order_equiv_sites(self, figure: Sequence[int]) -> List[int]:
        """Sort equivalent sites of a figure in index representation."""
        figure_cpy = list(figure)
        for eq_group in self.equiv_sites:
            equiv_indices = [figure[i] for i in eq_group]
            equiv_indices.sort()
            for count, i in enumerate(eq_group):
                figure_cpy[i] = equiv_indices[count]
        return figure_cpy

    @property
    def num_fig_occurences(self) -> Dict[str, int]:
        """Number of ocurrences for each figures."""
        occ_count = {}
        for figure in self.indices:
            key = self.get_figure_key(figure)
            current_num = occ_count.get(key, 0)
            occ_count[key] = current_num + 1
        return occ_count

    def corresponding_figure(self, ref_indx: int, target_figure: Sequence[int],
                             trans_matrix: List[Dict[int, int]]):
        """Find figures that correspond to another reference index.

        Parameters:

        ref_indx: int
            reference index

        target_figure: list of indices
            list of atomic indices that constitute the original figure before
            translating

        trans_matrix: list of dicts
            translation matrix
        """
        target_figure = self._order_equiv_sites(target_figure)
        for figure in self.indices:
            translated_figure = [trans_matrix[ref_indx][x] for x in figure]
            translated_figure = self._order_equiv_sites(translated_figure)
            if translated_figure == target_figure:
                return self._order_equiv_sites(figure)

        raise RuntimeError(f"There are no matching figure for ref_indx: "
                           f"{ref_indx} and figure: {target_figure}!")

    def get_all_figure_keys(self):
        return [self.get_figure_key(fig) for fig in self.indices]

    @property
    def multiplicity(self) -> int:
        return len(self.figures)



from typing import Sequence, List
import random
import numpy as np


class SwapMoveIndexTracker:

    def __init__(self):
        self.tracker = {}
        self.index_loc = None
        self._last_move = []

    @staticmethod
    def _unique_symbols_from_atoms(atoms) -> List[str]:
        return sorted(set(atoms.symbols))

    @property
    def symbols(self) -> List[str]:
        return list(self.tracker.keys())

    @property
    def num_symbols(self) -> int:
        return len(self.tracker)

    def __repr__(self):
        str_repr = f"SwapMoveIndexTracker at {hex(id(self))}\n"
        str_repr += f"Symbols tracked: {self.symbols}\n"
        str_repr += f"Tracker info: {self.tracker}\n"
        return str_repr

    def init_tracker(self, atoms):
        """Initialize the tracker with the numbers."""
        symbols = self._unique_symbols_from_atoms(atoms)

        # Track indices of all symbols
        self.tracker = {symb: [] for symb in symbols}

        # Track the location in self.tracker of each index
        self.index_loc = np.zeros(len(atoms), dtype=int)
        for atom in atoms:
            self.tracker[atom.symbol].append(atom.index)
            self.index_loc[atom.index] = len(self.tracker[atom.symbol]) - 1

    def filter_indices(self, include: Sequence[int]):
        include_set = set(include)
        for s, idx in self.tracker.items():
            self.tracker[s] = list(set(idx).intersection(include_set))
            for i, num in enumerate(self.tracker[s]):
                self.index_loc[num] = i

        # Remove empty items
        self.tracker = {k: v for k, v in self.tracker.items() if v}

    def move_already_updated(self, system_changes):
        """Return True if system_changes have already been taken into account.

        Parameters:

        system_changes:
        """
        return system_changes == self._last_move

    def update_swap_move(self, system_changes):
        """Update the atoms tracker.

        Parameters:

        system_changes:
        """
        if self.move_already_updated(system_changes):
            # This change has already been updated!
            return
        self._last_move = system_changes
        indx1 = system_changes[0][0]
        indx2 = system_changes[1][0]
        symb1 = system_changes[0][1]
        symb2 = system_changes[0][2]

        # Find the locations of the indices
        loc1 = self.index_loc[indx1]
        loc2 = self.index_loc[indx2]

        # Update the tracker and the locations
        self.tracker[symb1][loc1] = indx2
        self.index_loc[indx2] = loc1

        self.tracker[symb2][loc2] = indx1
        self.index_loc[indx1] = loc2

    def undo_last_swap_move(self):
        """Undo last swap move."""
        if not self._last_move:
            return
        opposite_change = []
        for change in self._last_move:
            opposite_change.append((change[0], change[2], change[1]))

        self.update_swap_move(opposite_change)
        self._last_move = []

    def get_random_indx_of_symbol(self, symbol: str) -> int:
        return random.choice(self.tracker[symbol])



# pylint: disable=undefined-variable
import logging
from deprecated import deprecated
from .version import *

# Do imports in the proper order, avoid cyclic imports
from . import datastructures
from . import basis_function
from . import logging_utils
from . import mp_logger
from . import cluster
from . import jsonio
from . import tools
from . import db_util
from . import corr_func
from . import settings
from . import calculator
from . import montecarlo
from . import structgen
from . import concentration

from .convexhull import *
from .data_manager import *
from .structure_mapper import *
from .structgen.new_struct import NewStructures  # Promotion for convenience
from .svd import *
from .concentration import *
from .regression_old import *
from .cluster_coverage import *
from .logging_utils import *
from .evaluate import *

# Import this global, so it can be disabled with "clease.REQUIRE_COMPATIBLE_TABLE_VERISON = False"
from .db_util import toggle_require_compatible_table_version

logger = logging.getLogger(__name__)


@deprecated(version='0.10.0', reason='import CEBulk from clease.settings instead')
def CEBulk(*args, **kwargs):
    return settings.CEBulk(*args, **kwargs)


@deprecated(version='0.10.0', reason='import CECrystal from clease.settings instead')
def CECrystal(*args, **kwargs):
    return settings.CECrystal(*args, **kwargs)


@deprecated(version='0.10.0', reason='import CESlab from clease.settings instead')
def CESlab(*args, **kwargs):
    return settings.CESlab(*args, **kwargs)


ADDITIONAL = ('__version__', 'version', 'settings', 'basis_function', 'jsonio', 'corr_func',
              'structgen', 'NewStructures', 'cluster', 'tools', 'montecarlo', 'mp_logger', 'CEBulk',
              'CECrystal', 'CESlab', 'concentration', 'db_util', 'calculator',
              'toggle_require_compatible_table_version', 'logging_utils', 'datastructures')

__all__ = (version.__all__ + concentration.__all__ + evaluate.__all__ + convexhull.__all__ +
           data_manager.__all__ + structure_mapper.__all__ + svd.__all__ + regression_old.__all__ +
           cluster_coverage.__all__ + logging_utils.__all__) + ADDITIONAL



from typing import Tuple
import numpy as np


class DataNormalizer:
    """
    Class for normalizing a data set to zero mean and unit variance

    Parameters:

    fail_on_constant: bool
        If True the normalizer will raise an error if a constant column
        is encountered. Otherwise, the standard deviation of these column
        will be set to 1.0
    """

    def __init__(self, fail_on_constant=False):
        self.meanX = None
        self.stdX = None
        self.meanY = None
        self.fail_on_constant = fail_on_constant

    def normalize(self, X: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Normalizes the each column of X to zero mean and unit variance. y is
        shifted to zero mean (variance is not altered)

        :param X: Design matrix shape (N x M)
        :param y: Target values length N

        Returns:
            X_norm, y_shifted (e.g. normalized X matrix and shifted y values)
        """
        self.meanY = np.mean(y)
        y_shifted = y - self.meanY
        self.meanX = np.mean(X, axis=0)
        self.stdX = np.std(X, axis=0, ddof=1)
        tol = 1e-16
        if np.any(self.stdX < tol) and self.fail_on_constant:
            raise ValueError(f"DataNormalizer: The following columns has a "
                             f"constant value:\n{self.constant_cols(X)}\n"
                             f"Please remove these.")
        self.stdX[self.stdX < tol] = 1.0
        X_norm = (X - self.meanX) / self.stdX
        return X_norm, y_shifted

    def constant_cols(self, X: np.ndarray, tol=1e-16) -> np.ndarray:
        """
        Return indices of columns in X that are constant

        Parameters:

        X: np.ndarray
            Design matrix for fitting

        tol: float
            Columns with a standard deviation smaller than this value are
            considered to be constant
        """
        # pylint: disable=no-self-use
        std = np.std(X, axis=0, ddof=1)
        return np.argwhere(std < tol)[:, 0]

    def varying_cols(self, X: np.ndarray, tol=1e-16) -> np.ndarray:
        """
        Return the indices of the columns that vary

        Parameters:

        X: np.ndarray
            Design matrix for fitting

        tol: float
            Columns with a standard deviation larger than this value are
            considered to be varying
        """
        # pylint: disable=no-self-use
        std = np.std(X, axis=0, ddof=1)
        return np.argwhere(std >= tol)[:, 0]

    def convert(self, coeff: np.ndarray) -> np.ndarray:
        """
        Converts coefficients obtained via X_norm.dot(coeff) = y_shifted to
        the ones that would have been obtained via X.dot(coeff) = y. Consult,
        the normalize function for an in depth explination of the difference
        between X_norm and X.

        Parameter:

        coeff: np.ndarray
            Array of coefficients obtained by using a normalized design matrix
            during fitting.

        Returns:
            converted coefficients
        """
        return coeff / self.stdX

    def bias(self, coeff: np.ndarray) -> float:
        """
        Return the bias term.
        """
        return self.meanY - np.sum(coeff * self.meanX / self.stdX)



import numpy as np
from .bias_potential import BiasPotential


class RandomBiasWithMemory(BiasPotential):
    """
    Adds a random normal distributed correction to each structure.
    Visited structures are remembered, such that a consistent correction is
    added to the structures

    Parameters:

    std: float
        Standard deviation of the correction in eV

    atoms: Atoms
        Atoms object used for the Monte Carlo sampling
    """

    def __init__(self, std, atoms):
        BiasPotential.__init__(self)
        self.std = std
        self.atoms = atoms
        self.history = {}

    def get_hash_key(self):
        """
        Return an integer hash corresponding to the current structure
        """
        return hash(tuple(self.atoms.numbers))

    def __call__(self, sytem_changes):
        """
        This method is called by the sampler after the changes of the trial
        move has been inserted. Thus, self.atoms is already updated at the
        time this function is called by the sampler.
        """
        hash_key = self.get_hash_key()
        value = self.history.get(hash_key, np.random.normal(scale=self.std))
        self.history[hash_key] = value
        return value

    def calculate_from_scratch(self, atoms):
        self.atoms.numbers = atoms.numbers
        return self(None)



import logging
import numpy as np
from matplotlib import pyplot as plt

logger = logging.getLogger(__name__)


class MCTrajectoryExtractor:
    """
    Class that extracts structures that are related by two atoms swap from
    a list of atoms
    """

    def is_related_by_swap(self, atoms1, atoms2):
        """
        Check if the second atom can be obtained by swapping two
        atoms
        """
        # pylint: disable=no-self-use
        if len(atoms1) != len(atoms2):
            return False

        differ1 = []
        differ2 = []
        for n1, n2 in zip(atoms1.numbers, atoms2.numbers):
            if n1 != n2:
                differ1.append(n1)
                differ2.append(n2)

        if len(differ1) != 2:
            return False

        return sorted(differ1) == sorted(differ2)

    def find_swaps(self, all_atoms):
        """
        Extract paths links all entries that are connected by swaps
        """
        swaps = []
        for i, atoms1 in enumerate(all_atoms):
            for j, atoms2 in enumerate(all_atoms[i + 1:]):
                if self.is_related_by_swap(atoms1, atoms2):
                    swaps.append((i, i + j + 1))
        return swaps

    def swap_energy_deviations(self, swaps, e_pred, e_ref):
        """
        Calculate the energy deviation between the energy difference in e_ref
        and the energy difference in e_pred

        Parameters:

        swaps: list
            List of tuples with swaps (e.g. [(0, 3), (2, 8)]). Can be obtained
            from find_swaps

        e_pred: list
            List of energies prediced for all energy calculations. Note that
            this should be total energies (not normalised per atom)

        e_ref: list
            List of energies from for instance DFT. Should be total energy, not
            normalised per atom.
        """
        # pylint: disable=no-self-use
        dev = []
        for s in swaps:
            dE_pred = e_pred[s[0]] - e_pred[s[1]]
            dE_ref = e_ref[s[0]] - e_ref[s[1]]
            dev.append(dE_pred - dE_ref)
        return dev

    def gaussian(self, x, mu, var):
        """
        Return the gaussian distribution

        Parameters:

        x: np.ndarray
            Values at which to evaluate the gaussian

        mu: float
            Expecteation value

        var: float
            Variance
        """
        # pylint: disable=no-self-use
        return np.exp(-(x - mu)**2 / (2 * var)) / np.sqrt(2 * np.pi * var)

    def plot_swap_deviation(self, dev):
        """
        Creates a plot with the distribution of the deviations
        """
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        hist, bins = np.histogram(np.array(dev), bins='auto')
        dx = bins[1] - bins[0]
        ax.plot(bins[1:] - 0.5 * dx, hist / np.trapz(hist, dx=dx), 'o', mfc='none', color='grey')
        ax.set_xlabel("Energy deviation (eV)")
        ax.set_ylabel("Probability Density")
        mu = np.mean(dev)
        var = np.var(dev)
        logger.info("Mean dev. %.4f eV", mu)
        logger.info("Standard deviation %.4f eV", np.sqrt(var))

        rng = bins[-1] - bins[0]
        x = np.linspace(bins[0] - 0.2 * rng, bins[-1] + 0.2 * rng, 100)
        y = self.gaussian(x, mu, var)
        ax.plot(x, y, color="#3c362b")
        x = np.linspace(mu - np.sqrt(var), mu + np.sqrt(var), 100)
        y = self.gaussian(x, mu, var)
        ax.fill_between(x, y, color="#8c361f", alpha=0.3)

        x = np.linspace(mu + np.sqrt(var), mu + 2 * np.sqrt(var), 100)
        y = self.gaussian(x, mu, var)
        ax.fill_between(x, y, color="#5b4930", alpha=0.3)

        x = np.linspace(mu - 2 * np.sqrt(var), mu - np.sqrt(var), 100)
        y = self.gaussian(x, mu, var)
        ax.fill_between(x, y, color="#5b4930", alpha=0.3)
        return fig
